package de.meinefirma.meinprojekt.backingbeans;

import java.lang.management.*;
import java.text.DecimalFormat;
import java.util.*;
import javax.management.*;

class CpuGcMemTxMBeanUtil
{
   private static final double NANOSEK_PRO_MILLISEK = 1000000.;
   private static final double MILLISEK_PRO_SEK     =    1000.;
   private static final double PROZENT_MAX_100      =     100.;
   private static final double SCHWELLE_NACHKOMMA   =      10.;

   /** Konvertiere double-Wert in String mit keiner oder einer Nachkommastelle. */
   static String doubleString1( double d )
   {
      final DecimalFormat df1 = new DecimalFormat( "0.0" );
      return ( d < SCHWELLE_NACHKOMMA ) ? df1.format( d ) : ("" + Math.round( d ));
   }

   /** Berechne prozentuale CPU-Zeit. */
   static int calculateCpuTimePercent(
         long cpuTimeNs, long lastCpuTimeNs, long upTimeMs, long lastUpTimeMs, long cpuCount )
   {
      return (int) Math.round( Math.min( PROZENT_MAX_100,
            (cpuTimeNs - lastCpuTimeNs) /
                  ((upTimeMs - lastUpTimeMs) * cpuCount * (NANOSEK_PRO_MILLISEK/PROZENT_MAX_100)) ) );
   }

   /** Berechne prozentuale Garbage-Collection-Zeit. */
   static double calculateGarbageCollectionTimePercent(
         long gcTimeMs, long lastGcTimeMs, long upTimeMs, long lastUpTimeMs, long cpuCount )
   {
      return Math.min( PROZENT_MAX_100, ((gcTimeMs - lastGcTimeMs) * PROZENT_MAX_100) /
            ((upTimeMs - lastUpTimeMs) * cpuCount) );
   }

   /** Berechne Anzahl pro Sekunde (z.B. Anzahl Transaktionen). */
   static String[] calculatePerSecond(
         long[] counts, long[] lastCounts, long upTimeMs, long lastUpTimeMs )
   {
      int n = Math.min( counts.length, lastCounts.length );
      String[] result = new String[n];
      for( int i = 0; i < n; i++ ) {
         result[i] = doubleString1( MILLISEK_PRO_SEK * (counts[i] - lastCounts[i]) / (upTimeMs - lastUpTimeMs) );
      }
      return result;
   }

   /** Lies Garbage-Collection-Zeit (Millisekunden). */
   static long readGarbageCollectionTimeMs( List<GarbageCollectorMXBean> gcMXBeans )
   {
      long gcTimeMs = 0;
      for( GarbageCollectorMXBean gc : gcMXBeans ) {
         gcTimeMs += gc.getCollectionTime();
      }
      return gcTimeMs;
   }

   /** Lies CPU-Zeit (Nanosekunden). */
   static long readCpuTimeNs( MBeanServerConnection mbs )
   {
      return readLongFromMBean( mbs, "java.lang:type=OperatingSystem", "ProcessCpuTime" );
   }

   /** Falls WebLogic: Lies Anzahl Transaktionen. */
   static long[] readTxCountsWebLogic( MBeanServerConnection mbs, String serverName )
   {
      long[] txValues = new long[2];
      txValues[0] = readLongFromMBean( mbs, "com.bea:ServerRuntime=" + serverName + ",Name=JTARuntime,Type=JTARuntime", "TransactionTotalCount" );
      txValues[1] = readLongFromMBean( mbs, "com.bea:ServerRuntime=" + serverName + ",Name=JTARuntime,Type=JTARuntime", "TransactionRolledBackTotalCount" );
      return txValues;
   }

   /** Lies Long-Attribut von MBean. */
   private static long readLongFromMBean( MBeanServerConnection mbs, String objectName, String attributeName )
   {
      Object attrObj = readAttrFromMBean( mbs, objectName, attributeName );
      if( attrObj instanceof Long ) {
         return ((Long) attrObj).longValue();
      }
      return -1;
   }

   /** Lies Double-Attribut von MBean. */
   private static double readDoubleFromMBean( MBeanServerConnection mbs, String objectName, String attributeName )
   {
      Object attrObj = readAttrFromMBean( mbs, objectName, attributeName );
      if( attrObj instanceof Double ) {
         return ((Double) attrObj).doubleValue();
      }
      return -1;
   }

   /** Lies Attribut von MBean. */
   private static Object readAttrFromMBean( MBeanServerConnection mbs, String objectName, String attributeName )
   {
      List<Object> lst = new ArrayList<>();
      try {
         Set<ObjectName> objectNames = mbs.queryNames( new ObjectName( objectName ), null );
         for( ObjectName on : objectNames ) {
            Object attr = mbs.getAttribute( on, attributeName );
            if( objectNames.size() == 1 ) {
               return attr;
            }
            lst.add( attr );
         }
         return lst;
      } catch( Exception ex ) {
         return null;
      }
   }

   /** Ermittle Garbage-Collector-MXBeans. */
   static List<GarbageCollectorMXBean> getGarbageCollectorMXBeans( MBeanServerConnection mbs )
   {
      try {
         List<GarbageCollectorMXBean> gcMXBeans = new ArrayList<>();
         ObjectName gcAllObjectName = new ObjectName( ManagementFactory.GARBAGE_COLLECTOR_MXBEAN_DOMAIN_TYPE + ",*" );
         Set<ObjectName> gcMXBeanObjectNames = mbs.queryNames( gcAllObjectName, null );
         for( ObjectName on : gcMXBeanObjectNames ) {
            GarbageCollectorMXBean gc = ManagementFactory.newPlatformMXBeanProxy(
                  mbs, on.getCanonicalName(), GarbageCollectorMXBean.class );
            gcMXBeans.add( gc );
         }
         return gcMXBeans;
      } catch( Exception ex ) {
         return null;
      }
   }
}
