package de.meinefirma.meinprojekt.backingbeans;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;

/** Inspiziere Libs und Klassen.
    Ermittelt im Classpath und im Web-Inf-Lib-Pfad vorhandene Libs, und in diesen Libs mehrfach vorhandene Klassen.
    Damit diese Utility-Klasse sowohl in einer JSF-Webapp als auch in einem Stand-alone-Kommandozeilentool
    verwendet werden kann, hat sie keine direkte Abhaengigkeit zu ServletContext,
    und zwei Konstruktoren fuer die beiden Einsatzzwecke. */
public class InspectLibsAndClasses
{
   // Die Eintraege im folgenden String-Array muessen an die gewuenschten Ausschluesse angepasst werden:
   private static final String[] EXCLD_ARR = new String[] {
      "about.html", "about.ini", "about.mappings", "about.properties", ".api_description", "build.properties",
      "eclipse32.gif", "eclipse32.png", "modeling32.png", ".options", "overview.html", "plugin.properties", "plugin.xml", "source_tips" };
   private static final Set<String> EXCLD_SET = new HashSet<>( Arrays.asList( EXCLD_ARR ) );
   private static final String RESOURCE_LIB_PATH = "/WEB-INF/lib";
   private Map<String,Set<String>> classesToJars = null;
   private List<String[]> libsAndMulClassCnt = null;
   private String ausgewaehlteLib = null;
   private ServletContextStellvertreter servletContext = null;

   /** Konstruktor falls ohne ServletContext (z.B. im Stand-alone-Kommandozeilentool) */
   public InspectLibsAndClasses()
   {
   }

   /** Konstruktor falls mit ServletContext (z.B. in JSF-Webapp, Beispiel siehe InspectImplementationsVersionsBean.java) */
   public InspectLibsAndClasses( ServletContextStellvertreter servletContext )
   {
      this.servletContext = servletContext;
   }

   /** Nur falls Stand-alone-Kommandozeilentool (nicht fuer Webapp) */
   public static void main( String[] args )
   {
      InspectLibsAndClasses inspectLibsAndClasses = new InspectLibsAndClasses();
      List<String[]> libsAndMulClassCnt = inspectLibsAndClasses.getLibsAndMulClassCnt();
      for( String[] libAndCount : libsAndMulClassCnt ) {
         if( libAndCount[1] != null && libAndCount[1].length() > 0 ) {
            Set<String> libsSet = new HashSet<>();
            inspectLibsAndClasses.setAusgewaehlteLib( libAndCount[0] );
            List<String[]> classAndLibsList = inspectLibsAndClasses.getMultipleClassesList();
            for( String[] classAndLibs : classAndLibsList ) {
               for( String lb : classAndLibs[1].split( "," ) ) {
                  libsSet.add( lb.trim() );
               }
            }
            int anz = inspectLibsAndClasses.anzahlKlassen( libAndCount[0] );
            String s = "ClasspathDublettenCheck: " + libAndCount[0] + " enthaelt " +
                       (( anz > 0 ) ? (anz + " relevante Dateien, davon ") : "") +
                       libAndCount[1] + " in mehreren Libs vorhandene Dateien " + libsSet + ".";
            System.out.println( s );
         }
      }
   }

   /** Liste aller Libs, und falls es in der jeweiligen Lib Klassen gibt, die es auch in anderen Libs gibt,
       jeweils die Anzahl solcher mehrfach vorkommenden Klassen
       (falls JsfInspectImplementationsVersions-JSF-Tool: verwendet in: inspectImplementationsVersions.xhtml) */
   public List<String[]> getLibsAndMulClassCnt()
   {
      if( libsAndMulClassCnt == null ) {
         Set<String> result = new TreeSet<>();
         classesToJars = new HashMap<>();
         List<URL> urls = getClassLoaderUrlList();
         for( URL url : urls ) {
            result.add( url.getPath() );
         }
         if( servletContext != null ) {
            Set<String> pathes = new TreeSet<>( servletContext.getResourcePaths( RESOURCE_LIB_PATH ) );
            result.addAll( pathes );
         }
         for( String path : result ) {
            visitFile( path, path );
         }
         libsAndMulClassCnt = collectMultipleClasses( result );
      }
      return libsAndMulClassCnt;
   }

   /** Zu einer einzelnen Lib die Liste der Klassen, welche auch in anderen Libs vorkommen,
       sowie die Auflistung der Libs, in der die Klasse vorkommt
       (falls JsfInspectImplementationsVersions-JSF-Tool: verwendet in: multipleClasses.xhtml) */
   public List<String[]> getMultipleClassesList()
   {
      if( ausgewaehlteLib == null ) { return null; }
      List<String[]> multipleClasses = new ArrayList<>();
      for( Map.Entry<String,Set<String>> me : classesToJars.entrySet() ) {
         Set<String> libs = me.getValue();
         if( libs.size() <= 1 || !libs.contains( extractLibName( ausgewaehlteLib ) ) ) { continue; }
         String lbStr = null;
         for( String lbi : libs ) {
            String lb = ( lbi.startsWith( "/" ) && (lbi.endsWith( ".jar" ) || lbi.endsWith( ".zip" )) )
                        ? lbi.substring( 1 ) : lbi;
            lbStr = concatWithComma( lbStr, "", lb );
         }
         multipleClasses.add( new String[] { me.getKey(), lbStr } );
      }
      return multipleClasses;
   }

   // Alternativen fuer Classloader bzw. Classpath (mit teilweise unterschiedlichen Ergebnissen):
   // ClassLoader cl = Thread.currentThread().getContextClassLoader();
   // ClassLoader cl = getClass().getClassLoader();
   // ClassLoader cl = ClassLoader.getSystemClassLoader();
   // String classpath = System.getProperty( "java.class.path" );
   private static List<URL> getClassLoaderUrlList()
   {
      List<URL> result = new ArrayList<>();
      ClassLoader cl = Thread.currentThread().getContextClassLoader();
      while( cl != null ) {
         if( cl instanceof URLClassLoader ) {
            URL[] urls = ((URLClassLoader) cl).getURLs();
            result.addAll( Arrays.asList( urls ) );
         }
         cl = cl.getParent();
      }
      return result;
   }

   private void visitFile( String rootPath, String fileOrDir )
   {
      boolean isRpDir = rootPath != null && (new File( rootPath )).isDirectory();
      File pf = new File( fileOrDir );
      if( pf.isDirectory() ) {
         File[] ff = pf.listFiles();
         if( ff != null ) {
            for( File f : ff ) {
               if( !isRpDir || !f.isDirectory() ) {
                  visitFile( rootPath, f.toString() );
               }
            }
         }
      } else if( fileOrDir.toLowerCase().endsWith( ".jar" ) || fileOrDir.toLowerCase().endsWith( ".zip" ) ) {
         if( !fileOrDir.endsWith( "jre/lib/rt.jar" ) ) {
            if( servletContext != null && fileOrDir.startsWith( RESOURCE_LIB_PATH ) ) {
               try( InputStream fin = servletContext.getResourceAsStream( fileOrDir ) ) {
                  visitJar( rootPath, fin );
               } catch( IOException e ) { /* ok */ }
            } else {
               try( InputStream fin = new FileInputStream( fileOrDir ) ) {
                  visitJar( rootPath, fin );
               } catch( IOException e ) { /* ok */ }
            }
         }
      } else {
         verarbeiteEinzelneDatei( rootPath, fileOrDir );
      }
   }

   private void visitJar( String rootPath, InputStream fin ) throws IOException
   {
      try( JarInputStream jin = new JarInputStream( fin ) ) {
         JarEntry jarEntry;
         while( (jarEntry = jin.getNextJarEntry()) != null ) {
            if( !jarEntry.isDirectory() && isDateiRelevant( jarEntry.getName() ) ) {
               verarbeiteEinzelneDatei( rootPath, jarEntry.getName() );
            }
         }
      }
   }

   private void verarbeiteEinzelneDatei( String rootPath, String datei )
   {
      Set<String> jarsToThisClass = classesToJars.get( datei );
      if( jarsToThisClass == null ) {
         jarsToThisClass = new HashSet<>();
      }
      jarsToThisClass.add( extractLibName( rootPath ) );
      classesToJars.put( datei, jarsToThisClass );
   }

   protected static boolean isDateiRelevant( String name )
   {
      String nam = name.toLowerCase( Locale.GERMAN );
      return !EXCLD_SET.contains( nam ) && !nam.startsWith( "about_files" ) &&
             !nam.startsWith( "icons/" ) && !nam.startsWith( "meta-inf" ) && !nam.endsWith( "/" );
   }

   private List<String[]> collectMultipleClasses( Set<String> libs )
   {
      Map<String,Integer> libToMulClsCnt = new HashMap<>();
      for( Map.Entry<String,Set<String>> me : classesToJars.entrySet() ) {
         if( me.getValue().size() <= 1 ) { continue; }
         for( String lb : me.getValue() ) {
            if( libToMulClsCnt.containsKey( lb ) ) {
               libToMulClsCnt.put( lb, Integer.valueOf( libToMulClsCnt.get( lb ).intValue() + 1 ) );
            } else {
               libToMulClsCnt.put( lb, Integer.valueOf( 1 ) );
            }
         }
      }
      List<String[]> result = new ArrayList<>();
      for( String lb : libs ) {
         Integer c = libToMulClsCnt.get( extractLibName( lb ) );
         if( c == null ) {
            result.add( new String[] { lb, "", "" } );
         } else {
            int anz = anzahlKlassen( lb );
            result.add( new String[] { lb, c.toString(), ( anz > 0 ) ? (" (von " + anz + ")") : "" } );
         }
      }
      return result;
   }

   private static String extractLibName( String path ) {
      if( path.toLowerCase().endsWith( ".jar" ) || path.toLowerCase().endsWith( ".zip" ) ) {
         int i = path.lastIndexOf( '/' );
         if( i >= 0 ) { return path.substring( i ); }
      }
      return path;
   }

   public int anzahlKlassen( String lib )
   {
      int result = 0;
      if( servletContext != null && lib.startsWith( RESOURCE_LIB_PATH ) ) {
         try( InputStream fin = servletContext.getResourceAsStream( lib ) ) {
            result += anzahlKlassen( fin );
         } catch( IOException e ) { /* ok */ }
      } else {
         try( InputStream fin = new FileInputStream( lib ) ) {
            result += anzahlKlassen( fin );
         } catch( IOException e ) { /* ok */ }
      }
      return result;
   }

   private int anzahlKlassen( InputStream fin )
   {
      int result = 0;
      try( JarInputStream jin = new JarInputStream( fin ) ) {
         JarEntry jarEntry;
         while( (jarEntry = jin.getNextJarEntry()) != null ) {
            if( !jarEntry.isDirectory() && isDateiRelevant( jarEntry.getName() ) ) {
               result++;
            }
         }
      } catch( IOException e ) { /* ok */ }
      return result;
   }

   public static String concatWithComma( String sPrev, String prefixForS2, Object s2 )
   {
      if( s2 == null || s2.toString().trim().length() == 0 ) { return sPrev; }
      return (( sPrev == null || sPrev.trim().length() == 0 ) ? "" : (sPrev + ", ")) + prefixForS2 + s2;
   }

   /** Falls JsfInspectImplementationsVersions-JSF-Tool:
       In inspectImplementationsVersions.xhtml wird hierueber die Lib fuer multipleClasses.xhtml ausgewaehlt */
   public void setAusgewaehlteLib( String ausgewaehlteLib )
   {
      this.ausgewaehlteLib = ausgewaehlteLib;
   }

   /** Interface um eine Abhaengigkeit zu javax.servlet.ServletContext zu vermeiden */
   interface ServletContextStellvertreter
   {
      InputStream getResourceAsStream( String resourcePath );
      Set<String> getResourcePaths( String resourceLibPath );
   }
}
